/*
 * mkThreads 1.2
 * -------------
 *
 * Please see the web pages for releases and documentation.
 *
 * Author: Michael Kraus
 *         mailto:mmg_kraus@csi.com
 *         http://ourworld.compuserve.com/homepages/mmg_kraus
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted.
 * The author makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless the author from any claims or
 * liability for loss arising out of such use.
 *
 */

/* required to built a dll using stubs. should be a compiler option */
/* #define USE_TCL_STUBS */

/* Required for tcl.h */
#define TCL_THREADS
/* Important: You MUST have a threads-enabled TCL/TK */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <tcl.h>

#ifndef TRUE
#  define TRUE  1
#  define FALSE 0
#endif

/* defines TCL_NEWTHREADAPI for Tcl version 8.3.1 or higher */
#if TCL_MAJOR_VERSION > 8
#  define MKT_NEWTHREADAPI
#else 
#  if TCL_MINOR_VERSION > 3
#    define MKT_NEWTHREADAPI
#  else
#    if TCL_MINOR_VERSION == 3
#      if TCL_RELEASE_SERIAL >= 1
#        define MKT_NEWTHREADAPI
#      endif
#    endif
#  endif
#endif

/* defines return type for _MktInitThread function */
#ifdef MKT_NEWTHREADAPI
#  define MKT_INITTHREADTYPE Tcl_ThreadCreateType
#else
#  define MKT_INITTHREADTYPE void
#endif

/* copied from sun's example.c */
#ifdef __WIN32__
#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN
#   if defined(_MSC_VER)
#       define EXPORT(a,b) __declspec(dllexport) a b
#       define DllEntryPoint DllMain
#   else
#       if defined(__BORLANDC__)
#           define EXPORT(a,b) a _export b
#       else
#           define EXPORT(a,b) a b
#       endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

EXTERN EXPORT(int,Mkthreads_Init) _ANSI_ARGS_((Tcl_Interp *));
EXTERN EXPORT(int,Mkthreads_SafeInit) _ANSI_ARGS_((Tcl_Interp *));
BOOL APIENTRY DllEntryPoint(HINSTANCE hInst, DWORD reason, LPVOID reserved)
{
  return TRUE;
}
#endif

/* mkThreads version number */
#define _VERSION           "1.2"

/* some acronyms for popular Tcl functions */
#define _NSO(pcText)       Tcl_NewStringObj( pcText, -1 )
#define _SSO(pO,pcText)    Tcl_SetStringObj( pO, pcText, -1 )
#define _GSO(pO)           Tcl_GetStringFromObj( pO, NULL )

#define _NIO(iVal)         Tcl_NewIntObj( iVal )
#define _SIO(pO,iVal)      Tcl_SetIntObj( pO, iVal )
#define _GIO(pO,piVal)     Tcl_GetIntFromObj( pI, pO, piVal )

#define _NBO(bVal)         Tcl_NewBooleanObj( bVal )
#define _SBO(pO,bVal)      Tcl_SetBooleanObj( pO, bVal )
#define _GBO(pO,pbVal)     Tcl_GetBooleanFromObj( pI, pO, pbVal )

#define _NDO(fVal)         Tcl_NewDoubleObj( fVal )
#define _SDO(pO,fVal)      Tcl_SetDoubleObj( pO, fVal )
#define _GDO(pO,pfVal)     Tcl_GetDoubleFromObj( pI, pO, pfVal )

#define _LOAL(pO,pNewO)    Tcl_ListObjAppendList( pI, pO, pNewO )
#define _LOAE(pO,pNewO)    Tcl_ListObjAppendElement( pI, pO, pNewO )
#define _LOGL(pO,piLen)    Tcl_ListObjLength( pI, pO, piLen )
#define _LOGI(pO,iI,poE)   Tcl_ListObjIndex( pI, pO, iI, poE )
#define _LOGE(pO,piC,ppV)  Tcl_ListObjGetElements( pI, pO, piC, ppV )
#define _LORE(pO,iPos,poE) Tcl_ListObjReplace( pI, pO, iPos, 1, 1, poE )
#define _LODE(pO,iPos)     Tcl_ListObjReplace( pI, pO, iPos, 1, 0, NULL )

#define _OSV(po1,po2,poV)  Tcl_ObjSetVar2( pI, po1, po2, poV, TCL_LEAVE_ERR_MSG )
#define _OGV(po1,po2)      Tcl_ObjGetVar2( pI, po1, po2, TCL_LEAVE_ERR_MSG )
#define _OUV(po1,po2)      Tcl_UnsetVar2( pI, _GSO(po1), (po2==NULL)?NULL:_GSO(po2), TCL_LEAVE_ERR_MSG )
#define _OSVG(po1,po2,poV) Tcl_ObjSetVar2( pI, po1, po2, poV, TCL_GLOBAL_ONLY )
#define _OGVG(po1,po2)     Tcl_ObjGetVar2( pI, po1, po2, TCL_GLOBAL_ONLY )
#define _OUVG(po1,po2)     Tcl_UnsetVar2( pI, _GSO(po1), (po2==NULL)?NULL:_GSO(po2), TCL_GLOBAL_ONLY )

#define _NOB               Tcl_NewObj()
#define _DOB               Tcl_DuplicateObj
#define _SOB(pO)           Tcl_IsShared( pO )? Tcl_DuplicateObj( pO ):pO
#define _ASO               Tcl_AppendStringsToObj

#define _DRC               Tcl_DecrRefCount
#define _IRC               Tcl_IncrRefCount

#define _GOR               Tcl_GetObjResult( pI )
#define _SOR(pO)           Tcl_SetObjResult( pI, pO )
#define _ROR               Tcl_ResetResult( pI )
#define _APR               Tcl_AppendResult

#define _HACE              Tcl_CreateHashEntry
#define _HADE              Tcl_DeleteHashEntry
#define _HASV              Tcl_SetHashValue
#define _HAFE              Tcl_FindHashEntry
#define _HAGV              Tcl_GetHashValue
#define _HAGK              Tcl_GetHashKey

#define _HAFESV(phH,pcK,psD,piN) _HASV( _HACE( phH, pcK, piN ), psD )
#define _HASCAN(phH,peE,psS)     peE = Tcl_FirstHashEntry( phH, psS ); peE != NULL; peE = Tcl_NextHashEntry( psS )

#define _GIFO(pO,pA,pcTxt,piRes) Tcl_GetIndexFromObj( pI, pO, pA, pcTxt, 0, piRes )
#define _WNA(objc,pcText)        ( Tcl_WrongNumArgs( pI, objc, objv, pcText ), TCL_ERROR )

#define try( Expr, Excep ) { if( (Expr) != TCL_OK ) throw Excep; }
#define throw              goto
#define catch

#define _MLOCK             Tcl_MutexLock(&sMutex)
#define _MUNLOCK           Tcl_MutexUnlock(&sMutex)

/* structs */

/* specifies a thread */
typedef struct Mkt_Thread {
  Tcl_ThreadId   hThread;                                  /* thread handle */
  Tcl_Interp    *psInterp;                           /* interpreter it uses */
  Tcl_HashEntry *psEntry;                     /* back pointer to hash table */
  Tcl_Condition  sWait;              /* used if it waits for the result ... */
  struct Mkt_Thread *psWfor;              /* of another thread, stored here */
  int            bRun;        /* TRUE normally. FALSE if we exit event loop */
} Mkt_Thread;

/* contains data passed to a new thread */
typedef struct Mkt_Starter {
  Tcl_Obj       *poScript;                  /* if not null, run this script */
  int            bInit;                           /* if true, call Tcl_Init */
} Mkt_Starter;

/* specifies a script sent to a thread */
typedef struct Mkt_Script {
  Mkt_Thread    *psThread;                   /* thread to execute script in */
  Tcl_Obj       *poScript;                                  /* a tcl script */
} Mkt_Script;

/* specifies an event for a thread's queue */
typedef struct Mkt_Event {
  Tcl_Event      sEvent;             /* must be here. see notifier man page */
  Mkt_Script    *psTarget;                  /* script and thread to execute */
  Mkt_Script    *psSource;            /* not null, if a callback is desired */
} Mkt_Event;

/* specifies a mutex/condition */
typedef struct Mkt_Mutex {
  Tcl_Obj       *poMutex;                              /* name of the mutex */
  Tcl_Obj       *poLockers;            /* list of threads trying to lock it */
  Tcl_Mutex      sMutex;                                   /* the tcl mutex */
  Tcl_Condition  sCondition;             /* ...with an integrated condition */
  Tcl_ThreadId   hOwner;           /* thread that currently holds the mutex */
  Tcl_ThreadId   hNotifier;                    /* thread that last notified */
  int            bTimeout;          /* true, if waiter experienced time-out */
} Mkt_Mutex;

/* modul global variables */
static Tcl_Mutex     sMutex;         /* to spawn threads and protect tables */
static Tcl_Condition sCond;     /* to spawn a thread and wait for its start */
static Tcl_HashTable tThreads;     /* threads. points to Mkt_Thread structs */
static Tcl_HashTable tMutexes;      /* mutexes. points to Mkt_Mutex structs */
static Tcl_HashTable tPackages;              /* packages to be auto-loaded. */
static Tcl_HashTable tProcs;                     /* procs to be auto-loaded */
static Tcl_HashTable tTvars;         /* thread variables. points to Tcl_Obj */
static Tcl_HashTable tSerials;  /* serialized sections. points to Tcl_Mutex */
static Mkt_Thread    sMainThread; /* the main thread. contained in tThreads */
static Mkt_Script    sOnexit;    /* script to execute when any thread exits */
static int           iMutexSeq;      /* counter for auto-naming new mutexes */
static Tcl_Obj      *poInitScript;  /* fixes problem with Tcl_Init in 8.3.1 */

/* the function prototypes. */
static void _MktCreateCommands( Mkt_Thread * );
static int  _MktGetTimeFromObj( Tcl_Interp *, Tcl_Obj *, Tcl_Time * );
static void _MktRemoveFromList( Tcl_Obj *, Mkt_Thread *, int * );
static void _MktAutoLoad      ( Mkt_Thread * );
static void _MktQueueEvent    ( Mkt_Script *, Mkt_Script * );
static void _MktExitThread    ( ClientData );
static MKT_INITTHREADTYPE _MktInitThread ( ClientData );
static int  _MktCreateThread  ( Tcl_Interp *, Tcl_Obj *, int );
static int  _MktOnReceive     ( Tcl_Event *, int );
static int  _MktSendScript    ( Tcl_Interp *, Mkt_Thread *, Tcl_Obj *, Mkt_Thread *, Tcl_Obj * );
static int  _MktOnExit        ( Tcl_Interp *, Mkt_Thread *, Tcl_Obj * );
static int  _MktSetPackages   ( Tcl_Interp *, Tcl_Obj * );
static int  _MktSetProcs      ( Tcl_Interp *, Tcl_Obj * );

int Mkt_ThreadCmd   ( ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[] );
int Mkt_MutexCmd    ( ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[] );
int Mkt_TvarCmd     ( ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[] );
int Mkt_SerializeCmd( ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[] );

int  Mkthreads_Init( Tcl_Interp * );
void Mkthreads_Exit( ClientData, Tcl_Interp * );

#ifndef MKT_NEWTHREADAPI
  extern int  TclpThreadCreate();
  extern void TclFinalizeMutex();
  extern void TclFinalizeCondition();
#endif

/* _MktCreateCommands
creates the new tcl commands of this package. called for the main thread
(obviously) and for each created thread. ClientData pC is a pointer to
the thread's respective Mkt_Thread struct, so no hash table lookup
will be required.
*/
static void _MktCreateCommands( Mkt_Thread *psThread )
{
  Tcl_Interp *pI = psThread->psInterp;
  ClientData pC = (ClientData)psThread;

  Tcl_CreateObjCommand( pI, "thread"   , Mkt_ThreadCmd   , pC, NULL );
  Tcl_CreateObjCommand( pI, "mutex"    , Mkt_MutexCmd    , pC, NULL );
  Tcl_CreateObjCommand( pI, "tvar"     , Mkt_TvarCmd     , pC, NULL );
  Tcl_CreateObjCommand( pI, "serialize", Mkt_SerializeCmd, pC, NULL );
}

/* _MktGetTimeFromObj
helper function to convert an integer contained in a Tcl_Obj
and representing milliseconds into a Tcl_Time struct.
*/
static int _MktGetTimeFromObj( Tcl_Interp *pI, Tcl_Obj *poTime, Tcl_Time *psTime )
{
  int iMsec;

  try( _GIO( poTime, &iMsec ), eError );

  if( iMsec < 0 )
    iMsec = 0;

  psTime->sec  = iMsec/1000;
  psTime->usec = (iMsec%1000)*1000;

  return TCL_OK;

  catch eError:
    return TCL_ERROR;
}

/* _MktRemoveFromList
helper function to remove a thread from a tcl list (see above). splits
up the list and searches through the elements by converting them into ints
and comparing it with the thread's id. if found, removes the element from.
the list. returns the index of where it was in piIndex, or -1 if not found.
*/
static void _MktRemoveFromList( Tcl_Obj *poList, Mkt_Thread *psThread, int *piIndex )
{
  int         iObjc, iElem;
  Tcl_Obj   **ppoObjv;

  Tcl_ListObjGetElements( NULL, poList, &iObjc, &ppoObjv );
  while( iObjc-- )
  {
    if( Tcl_GetIntFromObj( NULL, ppoObjv[iObjc], &iElem ) != TCL_OK ) continue;
    if( iElem != (int)psThread->hThread ) continue;
    Tcl_ListObjReplace( NULL, poList, iObjc, 1, 0, NULL );
    break;
  }

  if( piIndex )
    *piIndex = iObjc;
}

/* _MktAutoLoad
loads all packages and procs defined by 'thread procs' resp. 'thread packages'
into a thread's interpreter. called upon creation of a new thread. scans the
packages table whoes entries point to scripts returned by the "package
ifneeded" command and executes these scripts. then scans the procs table,
where the entries are executable scripts starting with "proc ...".
*/
static void _MktAutoLoad( Mkt_Thread *psThread )
{
  Tcl_Interp     *pI = psThread->psInterp;

  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;

  for( _HASCAN( &tPackages, psHashE, &sHashS ) )
    Tcl_EvalObj( pI, (Tcl_Obj*)_HAGV( psHashE ) );

  for( _HASCAN( &tProcs, psHashE, &sHashS ) )
    Tcl_EvalObj( pI, (Tcl_Obj*)_HAGV( psHashE ) );
}

/* _MktQueueEvent
queues an event into the queue of the thread specified in psTarget. simply
prepares an event structure (where the first element MUST be of type
Tcl_Event), queues it in and alerts the target thread (which will
call _MktOnReceive and then automatically free the memory. see the notifier
man page).
*/
static void _MktQueueEvent( Mkt_Script *psTarget, Mkt_Script *psSource )
{
  Mkt_Event *psEvent;

  psEvent = (Mkt_Event*)Tcl_Alloc( sizeof( Mkt_Event ) );
  memset( psEvent, 0, sizeof( Mkt_Event ) );

  psEvent->psTarget    = psTarget;
  psEvent->psSource    = psSource;
  psEvent->sEvent.proc = _MktOnReceive;

  Tcl_ThreadQueueEvent( psTarget->psThread->hThread, (Tcl_Event*)psEvent, TCL_QUEUE_TAIL );
  Tcl_ThreadAlert( psTarget->psThread->hThread );
}

/* _MktExitThread
exit handler for a terminating thread, which is passed in ClientData pC.
deletes the thread's associated entry in the hash table, deletes the thread's
interpreter and unlocks all pending mutexes. then searches the threads
table for any threads waiting on notification from this thread and notifies
them. then eventually queues the exit script, if defined.
*/
static void _MktExitThread( ClientData pC )
{
  Tcl_Interp *pI = NULL;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;
  Mkt_Mutex  *psMutex;
  Mkt_Thread *psWaiter, *psThread = (Mkt_Thread*)pC;
  Mkt_Script *psScript;

  _MLOCK;

  /* general clean-up */
  _HADE( psThread->psEntry );
  Tcl_DeleteInterp( psThread->psInterp );

  /* unlock any mutexes this thread is holding */
  for( _HASCAN( &tMutexes, psHashE, &sHashS ) )
  {
    psMutex = (Mkt_Mutex*)_HAGV( psHashE );
    _MktRemoveFromList( psMutex->poLockers, psThread, NULL );
    if( psMutex->hOwner != psThread->hThread ) continue;

    psMutex->hOwner = 0;
    Tcl_MutexUnlock( &psMutex->sMutex );
  }

  /* look for waiting threads and notify them */
  for( _HASCAN( &tThreads, psHashE, &sHashS ) )
  {
    psWaiter = (Mkt_Thread*)_HAGV( psHashE );
    if( psWaiter->psWfor == psThread )
      Tcl_ConditionNotify( &psWaiter->sWait );
  }

  /* eventually trigger the exit script */
  if( sOnexit.poScript )
  {              /* create physical copy of script and append thread handle */
    psScript = (Mkt_Script*)Tcl_Alloc( sizeof( Mkt_Script ) );
    psScript->psThread = sOnexit.psThread;
    psScript->poScript = _DOB( sOnexit.poScript );
    _LOAE( psScript->poScript, _NIO( (int)psThread->hThread ) );
    _IRC( psScript->poScript );
    _MktQueueEvent( psScript, NULL );       /* _MktOnReceive will call free */
  }

  _MUNLOCK;

  /* unfortunately conditions leak memory, even with clean-up! */
#ifdef MKT_NEWTHREADAPI
  if( psThread->sWait )
    Tcl_ConditionFinalize( &psThread->sWait );
#else
  if( psThread->sWait )
    TclFinalizeCondition( &psThread->sWait );
#endif

  Tcl_Free( (char*)psThread );
}

/* _MktInitThread
initializes a new thread. a script to execute is passed in ClientData pC.
creates and initializes a record for the threads table, hereby creating
the thread's associated interpreter. then notifies the creating thread,
which is waiting on a condition, that it can continue. at this time, the
script passed by the creating thread has been saved and the threads table
modification is done. then defines an exit handler, creates the commands
of this package for the new interpreter and autoloads all defined packages
and procs. finally, if a script has been specified, evaluates it and
terminates the thread. if not, it waits on the event loop to receive
scripts sent from other threads via 'thread eval'.
*/
static MKT_INITTHREADTYPE _MktInitThread( ClientData pC )
{
  Mkt_Starter *psStarter = (Mkt_Starter*)pC;

  int         iRes, bNew, bInit;
  Tcl_Obj    *poScript;
  Mkt_Thread *psThread;

  /* that's from the creating thread's stack. save it! */
  bInit    = psStarter->bInit;
  poScript = psStarter->poScript;
  if( poScript )
    _IRC( poScript );

  _MLOCK;

  /* create a thread structure */
  psThread = (Mkt_Thread*)Tcl_Alloc( sizeof(Mkt_Thread) );
  memset( psThread, 0, sizeof(Mkt_Thread) );

  psThread->bRun     = TRUE;
  psThread->hThread  = Tcl_GetCurrentThread();
  psThread->psInterp = Tcl_CreateInterp();
  psThread->psEntry  = _HACE( &(tThreads), (char*)psThread->hThread, &bNew );
  _HASV( psThread->psEntry, psThread );

  /* register callback function to release resources */
  Tcl_CreateThreadExitHandler( _MktExitThread, (ClientData)psThread );

  /* if requested by creating thread, call Tcl_Init to source init.tcl etc. */
#ifdef MKT_NEWTHREADAPI
  if( bInit )
    Tcl_EvalObj( psThread->psInterp, poInitScript );
#else
  if( bInit )
    Tcl_Init( psThread->psInterp );
#endif

  /* do the rest of the initialization */
  _MktCreateCommands( psThread );
  _MktAutoLoad( psThread );

  /* now we can let the creating thread go */
  Tcl_ConditionNotify( &sCond );
  _MUNLOCK;

  if( poScript )
  {
    /* either execute any passed script ... */
    iRes = Tcl_EvalObj( psThread->psInterp, poScript );
    _DRC( poScript );
  }
  else 
  {
    /* ... or wait for scripts sent from other threads */    
    iRes = TCL_OK;
    while( psThread->bRun )
      Tcl_DoOneEvent( TCL_ALL_EVENTS );
  }

  Tcl_ExitThread( iRes );

#ifdef MKT_NEWTHREADAPI
  TCL_THREAD_CREATE_RETURN;
#endif
}

/* _MktCreateThread
spawns a new thread and eventually passes it a script. waits for notification
on the condition until the new thread doesn't need the stack data poScript
anymore. sets the result to the new thread's handle and exits.
*/
static int _MktCreateThread( Tcl_Interp *pI, Tcl_Obj *poScript, int bInit )
{
  Tcl_ThreadId hThread;
  Mkt_Starter  sStarter;

  _MLOCK;

  /* set struct with startup script and init flag */
  sStarter.poScript  = poScript;
  sStarter.bInit     = bInit;

  /* spawn thread which will execute _MktInitThread */
#ifdef MKT_NEWTHREADAPI
  try( Tcl_CreateThread( &hThread, _MktInitThread, (ClientData)&sStarter, TCL_THREAD_STACK_DEFAULT, TCL_THREAD_NOFLAGS ), eError );
#else
  try( TclpThreadCreate( &hThread, _MktInitThread, &sStarter ), eError );
#endif

  /* wait for alive signal of the new thread */
  Tcl_ConditionWait( &sCond, &sMutex, NULL );
  _MUNLOCK;

  _SOR( _NIO( (int)hThread ) );

  return TCL_OK;

  catch eError:
    _MUNLOCK;
    _SSO( _GOR, "Cannot create thread." );
    return TCL_ERROR;
}

/* _MktOnReceive
event handler for a thread. psTclEvent is really of type Mkt_Event. It
contains a script for the thread psTarget, which it executes. if psSource
is not null, then either the thread that sent the script is waiting for
this thread to return the result directly, or it wants an own script to be
called. in the latter case, the script in psSource is not null.
*/
static int _MktOnReceive( Tcl_Event *psTclEvent, int iMask )
{
  int iCode;
  Mkt_Event  *psEvent  = (Mkt_Event*)psTclEvent;
  Mkt_Script *psTarget = psEvent->psTarget;
  Mkt_Script *psSource = psEvent->psSource;
  Mkt_Thread *psThread = psTarget->psThread;
  Tcl_Interp *pI       = psThread->psInterp;
  Tcl_Obj    *poRet, *poScript;

  /* preserve the interpreter, if the script causes the thread to exit */
  Tcl_Preserve( (ClientData)pI );
  iCode = Tcl_EvalObj( pI, psTarget->poScript );         /* eval the script */

  /* if some feedback was desired by the sender */
  if( psSource )
  {
    if( iCode == TCL_ERROR && ! psThread->bRun )  /* we had a 'thread exit' */
      iCode = TCL_OK;                    /* ...which is not really an error */

    poRet = _NOB;                                         /* prepare a list */
    _LOAE( poRet, _NIO( iCode ) );                   /* get the return code */
    _LOAE( poRet, _GOR );                          /* get the result string */

    if( iCode == TCL_ERROR )                    /* if there was an error... */
    {                                    /* then also get error information */
      _LOAE( poRet, Tcl_GetVar2Ex( pI, "errorCode", NULL, TCL_GLOBAL_ONLY ) );
      _LOAE( poRet, Tcl_GetVar2Ex( pI, "errorInfo", NULL, TCL_GLOBAL_ONLY ) );
    }

    if( psSource->poScript )          /* here the sender defined a callback */
    {
      poScript = _DOB( psSource->poScript ); /* create a copy of the script */
      _LOAL( poScript, poRet );               /* to which we can append now */
      _IRC( poScript );                       /* ...because it's not shared */
      _DRC( psSource->poScript );                    /* forget the original */
      _DRC( poRet );
      psSource->poScript = poScript;               /* let point to the copy */
      _MktQueueEvent( psSource, NULL );    /* queue it back into the sender */
    }                       /* ...who will receive it in this very function */
    else                                      /* here the sender is waiting */
    {                                                   /* store the result */
      psSource->poScript = poRet; _IRC( psSource->poScript );
      _MLOCK;                    /* and notify the sender, who will free it */
      Tcl_ConditionNotify( &psSource->psThread->sWait );
      _MUNLOCK;
    }
  }

  Tcl_Release( (ClientData)pI );                 /* release the interpreter */
  _DRC( psTarget->poScript );         /* clean up: delete the script object */
  Tcl_Free( (char*)psTarget );               /* ... and the psTarget struct */
                                   /* (note: psSource freed by the sender!) */
  return 1;                         /* this will free psEvent automatically */
}

/* _MktSendScript
sends a script to another thread via the event queue. the receiving thread
is called with _MktOnReceive, where the script is evaluated. the target thread
is in psThread, the script in poScript. if this thread wants to get informed
about the result, then psOrig points to it. if this thread waits for the
target thread, then poCallback is null. if it wants to get notified
asynchronuosly by the target thread then poCallback contains a script, which
is put by the target thread into this thread's queue, along with result
information.
*/
static int _MktSendScript( Tcl_Interp *pI, Mkt_Thread *psThread, Tcl_Obj *poScript, Mkt_Thread *psOrig, Tcl_Obj *poCallback )
{
  int          iCode, iObjc;
  Tcl_Obj    **ppoObjv;
  Mkt_Script  *psSource, *psTarget;

  /* specify the target by means of a Mkt_Script struct */
  psTarget = (Mkt_Script*)Tcl_Alloc( sizeof( Mkt_Script ) );
  psTarget->poScript = poScript; _IRC( psTarget->poScript );
  psTarget->psThread = psThread;

  if( psOrig )             /* if some feedback is requested from the target */
  {                                                  /* ... then specify it */
    psSource = (Mkt_Script*)Tcl_Alloc( sizeof( Mkt_Script ) );
    psSource->poScript = poCallback;
    psSource->psThread = psOrig;

    if( psSource->poScript )                  /* for asynchronuous callback */
      _IRC( psSource->poScript );                    /* ... save the script */
    else                                       /* for waiting on the result */
    {
      _MLOCK;                          /* hold the target until we're ready */
      psSource->psThread->psWfor = psThread;     /* store the target thread */
    }                /* ...so we get unblocked when the thread should exit. */
  }                           /* otherwise we wouldn't need that reference. */
  else
    psSource = NULL;

  _MktQueueEvent( psTarget, psSource );

  if( poCallback || ! psOrig )               /* return if we're not waiting */
    return TCL_OK;

  Tcl_ConditionWait( &psOrig->sWait, &sMutex, NULL );         /* let's wait */
  _MUNLOCK;

  if( psSource->poScript )            /* the target thread stored something */
  {           /* and we expect it to be code, result, and maybe error infos */
    try( _LOGE( psSource->poScript, &iObjc, &ppoObjv ), eError );
    try( _GIO( ppoObjv[0], &iCode ), eError );
    _SOR( ppoObjv[1] );           /* make the target's result to our result */

    if( iObjc > 2 )
    {                                  /* ...and also any error information */
      Tcl_SetObjErrorCode( pI, ppoObjv[2] );
      Tcl_AddObjErrorInfo( pI, _GSO( ppoObjv[3] ), -1 );
    }

    _DRC( psSource->poScript );                        /* free that storage */
  }
  else                                  /* here the target must have exited */
    iCode = 0;

  psSource->psThread->psWfor = NULL;
  Tcl_Free( (char*)psSource );                               /* free source */
                                         /* (target freed by _MktOnReceive) */
  return iCode;                          /* mirror the target's return code */

  catch eError:
    return TCL_ERROR;
}

/* _MktOnExit
retrieves or sets an exit script for exiting threads. any thread can call this,
but there's always only one exit script. a thread's exit handler would append
its own thread handle as an argument to the script and queue it into this
thread. if poScript is null, the current settings are returned as a list.
*/
static int _MktOnExit( Tcl_Interp *pI, Mkt_Thread *psThread, Tcl_Obj *poScript )
{
  _MLOCK;

  if( poScript )                                 /* a new script is defined */
  {
    if( sOnexit.poScript )                /* if a script was defined before */
      _DRC( sOnexit.poScript );                       /* ... then forget it */

    if( strlen( _GSO( poScript ) ) )  /* empty string means: no exit script */
    {
      sOnexit.poScript = poScript; _IRC( sOnexit.poScript );
      sOnexit.psThread = psThread;            /* ...store script and thread */
    }
  }
  else if( sOnexit.poScript )      /* we should return the current settings */
  {
    _SOR( _NIO( (int)sOnexit.psThread->hThread ) );   /* they are thread id */
    try( _LOAE( _GOR, sOnexit.poScript ), eError );    /* ...and the script */
  }

  _MUNLOCK;

  return TCL_OK;

  catch eError:
    _MUNLOCK;
    return TCL_ERROR;
}

/* _MktSetPackages
sets or returns a list of packages which are automatically loaded upon a
thread's initialization. unfortunately, the suitable API function
Tcl_PkgRequire() seems to leak memory every time a thread terminates.
therefore we use "package ifneeded" to get the script that would actually
load the package, and rather execute the script later.
*/
static int _MktSetPackages( Tcl_Interp *pI, Tcl_Obj *poPackages )
{
  int             i, iObjc, bNew;
  char           *pcVersion;
  Tcl_Obj       **ppoObjv;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;

  _MLOCK;

  if( poPackages )                            /* a new list has been passed */
  {
    for( _HASCAN( &tPackages, psHashE, &sHashS ) )      /* forget about all */
      _DRC( (Tcl_Obj*)_HAGV( psHashE ) );                 /* ...old entries */

    Tcl_DeleteHashTable( &tPackages );
    Tcl_InitHashTable( &tPackages, TCL_STRING_KEYS );

    try( _LOGE( poPackages, &iObjc, &ppoObjv ), eError );
    for( i = 0; i < iObjc; i++ )
    {
      if( ( pcVersion = Tcl_PkgPresent( pI, _GSO( ppoObjv[i] ), NULL, 0 ) ) == NULL )
        throw eError;                 /* complain if package is not present */
                                 /* get and store the package's load script */
      try( Tcl_VarEval( pI, "package ifneeded ", _GSO( ppoObjv[i] ), " ", pcVersion, NULL ), eError );
      _HAFESV( &tPackages, _GSO( ppoObjv[i] ), _GOR, &bNew );
      _IRC( _GOR );
    }

    _ROR;
  }
  else                                       /* return current package list */
  {                           /* ... by appending all entries to the result */
    for( _HASCAN( &tPackages, psHashE, &sHashS ) )
      _LOAE( _GOR, _NSO( _HAGK( &tPackages, psHashE ) ) );
  }

  _MUNLOCK;

  return TCL_OK;

  catch eError:
    _MUNLOCK;
    return TCL_ERROR;
}

/* _MktSetProcs
sets or returns a list of procs which are automatically defined upon a
thread's initialization. each hash table entry will point to a complete
specification suitable for the proc command (name args body). the "info"
command is used to retrieve the elements of a procedure: the body, all
arguments and any default argument values.
*/
static int _MktSetProcs( Tcl_Interp *pI, Tcl_Obj *poProcs )
{
  int             i, j, iObjc, bNew, iParc;
  Tcl_Obj       **ppoObjv, *poPars, *poCall, **ppoParv, *poArgs, *poArg;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;

  _MLOCK;

  if( poProcs )                                    /* a new list was passed */
  {
    for( _HASCAN( &tProcs, psHashE, &sHashS ) )       /* forget old entries */
      _DRC( (Tcl_Obj*)_HAGV( psHashE ) );

    Tcl_DeleteHashTable( &tProcs );
    Tcl_InitHashTable( &tProcs, TCL_STRING_KEYS );

    try( _LOGE( poProcs, &iObjc, &ppoObjv ), eError );
    for( i = 0; i < iObjc; i++ )
    {
      poArgs = _NOB;           /* prepare a list for the args with defaults */
      try( Tcl_VarEval( pI, "info args ", _GSO( ppoObjv[i] ), NULL ), eError );
      poPars = _GOR; _IRC( poPars );     /* list of parameters, but without */
      try( _LOGE( poPars, &iParc, &ppoParv ), eError );  /* ...the defaults */

      for( j = 0; j < iParc; j++ )    /* for each parameter get any default */
      {
        poArg = _SOB( ppoParv[j] );                   /* create a parameter */
        try( Tcl_VarEval( pI, "info default ", _GSO( ppoObjv[i] ), " ", _GSO( poArg ), " _mkt_", NULL ), eError );

        if( ! strcmp( _GSO( _GOR ), "1" ) )        /* ...append any default */
          _LOAE( poArg, _NSO( Tcl_GetVar( pI, "_mkt_", 0 ) ) );

        _LOAE( poArgs, poArg );     /* ...and append it to the list of args */
      }

      _DRC( poPars );                                 /* forget temp object */

      poCall = _NSO( "proc" ); _IRC( poCall );      /* prepare call to proc */
      _LOAE( poCall, ppoObjv[i] );                   /* ...append proc name */
      _LOAE( poCall, poArgs );                          /* ...the args list */

      try( Tcl_VarEval( pI, "info body ", _GSO( ppoObjv[i] ), NULL ), eError );
      _LOAE( poCall, _GOR );                            /* ... and the body */

      _HAFESV( &tProcs, _GSO( ppoObjv[i] ), poCall, &bNew );   /* store it! */
    }

    _ROR;
  }
  else                                         /* return current procs list */
  {                           /* ... by appending all entries to the result */
    for( _HASCAN( &tProcs, psHashE, &sHashS ) )
      try( _LOAE( _GOR, _NSO( _HAGK( &tProcs, psHashE ) ) ), eError );
  }

  _MUNLOCK;

  return TCL_OK;

  catch eError:
    _MUNLOCK;
    return TCL_ERROR;
}

/* Mkt_ThreadCmd
implements the thread command. it does several things, depending on the
specified option. a thread in this package is specified via an Mkt_Thread
struct. it stores the thread's handle, a pointer to the thread's interpreter
(each thread must have an own one, see the threads man page), a pointer to
its own hash table entry (as a shortcut to avoid look-ups), and information
if this thread is currently waiting on another thread to evaluate a script
that has been sent by this thread. the latter is only needed, when the other
tread wants to exit and therefore needs to unblock any waiting thread(s).
*/
int Mkt_ThreadCmd( ClientData pC, Tcl_Interp *pI, int objc, Tcl_Obj *CONST objv[] )
{
  Mkt_Thread *psThread = (Mkt_Thread *)pC;

  int             iOption, bWait, bInit;
  Tcl_ThreadId    hThread;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;
  Tcl_Obj        *poScript;
  Tcl_Channel     hChannel;
  Mkt_Thread     *psTarget;

  static char *pArg[] = { "create", "id", "eval", "exit", "names", "packages",
                          "procs", "exists", "onexit", "_register_", "_unregister_", NULL };

  enum options { _CREATE, _ID, _EVAL, _EXIT, _NAMES, _PACKAGES,
                 _PROCS, _EXISTS, _ONEXIT, _REGISTER, _UNREGISTER };

  if( objc < 2 )
    return _WNA( 1, "option ?arg?" );

  try( _GIFO( objv[1], pArg, "option", &iOption ), eError );

  switch( iOption )
  {
    case _CREATE:                                    /* create a new thread */
      bInit = ( objc > 2 && ! strcmp( _GSO( objv[2] ), "-init" ) );

      if( objc > 4 || ( objc == 4 && ! bInit ) )
        return _WNA( 2, "?-init? ?script?" );

      switch( objc )
      {
        case 2: poScript = NULL;                      break;
        case 3: poScript = ( bInit )? NULL : objv[2]; break;
        case 4: poScript = objv[3];                   break;
      }

      try( _MktCreateThread( pI, poScript, bInit ), eError );
      break;

    case _EVAL:                                /* send a script to a thread */
      if( objc < 4 || objc > 5 )
        return _WNA( 2, "?-wait? threadId script ?command?" );

      bWait = ( objc == 5 && ! strcmp( _GSO( objv[2] ), "-wait" ) );
      try( _GIO( bWait?objv[3]:objv[2], (int*)&hThread ), eError );
      poScript = bWait?objv[4]:objv[3];

      if( ( psHashE = _HAFE( &tThreads, (char*)hThread ) ) == NULL )
        throw eBadThread;                /* check if thread is in the table */

      psTarget = (Mkt_Thread*)_HAGV( psHashE );
      if( bWait )                                  /* this thread will wait */
        _MktSendScript( pI, psTarget, poScript, psThread, NULL );
      else if( objc == 5 )         /* will not wait but wants to get called */
        _MktSendScript( pI, psTarget, poScript, psThread, objv[4] );
      else                     /* will neither wait nor wants to get called */
        _MktSendScript( pI, psTarget, poScript, NULL, NULL );

      break;

    case _ONEXIT:                                  /* define an exit script */
      if( objc > 3 )
        return _WNA( 2, "?command?" );

      try( _MktOnExit( pI, psThread, (objc==3)?objv[2]:NULL ), eError );
      break;

    case _PACKAGES:                         /* define packages to auto-load */
      if( objc > 3 )
        return _WNA( 2, "?packageList?" );

      try( _MktSetPackages( pI, (objc==3)?objv[2]:NULL ), eError );
      break;

    case _PROCS:                          /* define procedures to auto-load */
      if( objc > 3 )
        return _WNA( 2, "?procList?" );

      try( _MktSetProcs( pI, (objc==3)?objv[2]:NULL ), eError );
      break;

    case _REGISTER:                 /* "share" a channel. very dangerous!!! */
      if( objc != 4 )
        return _WNA( 2, "channelId threadId" );

      hChannel = Tcl_GetChannel( psThread->psInterp, _GSO( objv[2] ), NULL );
      if( hChannel == NULL )
        throw eBadChannel;

      try( _GIO( objv[3], (int*)&hThread ), eError );      /* target thread */
      if( ( psHashE = _HAFE( &tThreads, (char*)hThread ) ) == NULL )
        throw eBadTargetThread;          /* check if thread is in the table */

      psTarget = (Mkt_Thread*)_HAGV( psHashE );
      Tcl_RegisterChannel( psTarget->psInterp, hChannel );

      break;

    case _UNREGISTER:        /* "unshare" a channel. even more dangerous!!! */
      if( objc < 3 || objc > 4 )
        return _WNA( 2, "channelId ?threadId?" );

      hChannel = Tcl_GetChannel( psThread->psInterp, _GSO( objv[2] ), NULL );
      if( hChannel == NULL )
        throw eBadChannel;

      if( objc == 4 )                       /* unregister from other thread */
      {
        try( _GIO( objv[3], (int*)&hThread ), eError );    /* target thread */
        if( ( psHashE = _HAFE( &tThreads, (char*)hThread ) ) == NULL )
          throw eBadTargetThread;        /* check if thread is in the table */

        psTarget = (Mkt_Thread*)_HAGV( psHashE );
        Tcl_UnregisterChannel( psTarget->psInterp, hChannel );
      }
      else                                   /* unregister from this thread */
        Tcl_UnregisterChannel( psThread->psInterp, hChannel );

      break;

    case _ID:                               /* return handle of this thread */
      if( objc > 2 )
        return _WNA( 2, "" );

      _SOR( _NIO( (int)psThread->hThread ) );
      break;

    case _EXIT:                                         /* exit this thread */
      psThread->bRun = FALSE;                /* set flag to exit event loop */
      Tcl_ThreadAlert( psThread->hThread );           /* wake up the thread */
      return TCL_ERROR;         /* required to fully unwind from call stack */

    case _EXISTS:                          /* test if a given thread exists */
      if( objc != 3 )
        return _WNA( 2, "threadId" );

      _MLOCK;                 /* simply checks if it's in the threads table */
      try( _GIO( objv[2], (int*)&hThread ), eError );
      _SOR( _NIO( _HAFE( &tThreads, (char*)hThread ) != NULL ) );
      _MUNLOCK;
      break;

    case _NAMES:                           /* return handles of all threads */
      if( objc > 2 )
        return _WNA( 2, "" );

      _MLOCK;             /* appends all entries in the table to the result */
      for( _HASCAN( &tThreads, psHashE, &sHashS ) )
        _LOAE( _GOR, _NIO( (int)_HAGK( &tThreads, psHashE ) ) );

      _MUNLOCK;
      break;
  }

  return TCL_OK;

  catch eError:
    return TCL_ERROR;

  catch eBadThread:
    _APR( pI, "No such thread ", _GSO( objv[objc-2] ), NULL );
    return TCL_ERROR;

  catch eBadTargetThread:
    _APR( pI, "No such thread ", _GSO( objv[3] ), NULL );
    return TCL_ERROR;

  catch eBadChannel:
    _APR( pI, "No such channel ", _GSO( objv[2] ), NULL );
    return TCL_ERROR;
}

/* Mkt_MutexCmd
implements the mutex command. does several things, depending on the
specified option. a mutex in this package does not only consist of a
Tcl_Mutex but also comes with a Tcl_Condition on which a thread can wait.
see the threads man page for how they work together. Mkt_Mutex also maintains
a list of threads that try to aquire it. It uses that list to prevent the
same thread from locking or unlocking a mutex twice (which can have undefined
results due to the TCL documentation). also it keeps information about who
owns the mutex, who notified the condition and if there was a timeout while
waiting for it (which is not a feature of a Tcl_Condition).
*/
int Mkt_MutexCmd( ClientData pC, Tcl_Interp *pI, int objc, Tcl_Obj *CONST objv[] )
{
  Mkt_Thread *psThread = (Mkt_Thread *)pC;

  int             iOption, bNew, iIndex;
  char            pcBuf[20], *pcMutex = pcBuf;
  Tcl_Time        sTime, *psTime = &sTime;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;
  Mkt_Mutex      *psMutex;

  static char *pArg[] = { "create", "delete", "lock", "unlock", "lockers", "owner",
                          "names", "wait", "notify", "notifier", "exists", NULL };

  enum options { _CREATE, _DELETE, _LOCK, _UNLOCK, _LOCKERS, _OWNER,
                 _NAMES, _WAIT, _NOTIFY, _NOTIFIER, _EXISTS };

  if( objc < 2 )
    return _WNA( 1, "option ?arg?" );

  try( _GIFO( objv[1], pArg, "option", &iOption ), eError );

  switch( iOption )                     /* do some option checking up-front */
  {
    case _CREATE: if( objc < 2 || objc > 3 ) return _WNA( 2, "?name?" )     ; break;
    case _WAIT  : if( objc < 3 || objc > 4 ) return _WNA( 2, "name ?msec?" ); break;
    case _NAMES : if( objc != 2 )            return _WNA( 2, "" )           ; break;
    default     : if( objc != 3 )            return _WNA( 2, "name" )       ; break;
  }

  _MLOCK;

  if( iOption != _CREATE && iOption != _NAMES && iOption != _EXISTS )
  {
    if( ( psHashE = _HAFE( &tMutexes, _GSO( objv[2]) ) ) == NULL )
      throw eBadMutex;

    psMutex = (Mkt_Mutex*)_HAGV( psHashE );         /* get pointer to mutex */
  }

  switch( iOption )
  {
    case _CREATE:                                         /* create a mutex */
      if( objc == 2 )                 /* generate a name, if none specified */
        sprintf( pcMutex, "mutex%d", ++iMutexSeq );
      else
        pcMutex = _GSO( objv[2] );

      psHashE = _HACE( &tMutexes, pcMutex, &bNew );         /* create entry */
      if( ! bNew )                          /* ...exit if is exists already */
        throw eMutexExists;

      psMutex = (Mkt_Mutex*)Tcl_Alloc( sizeof(Mkt_Mutex) );    /* init struct */
      memset( psMutex, 0, sizeof(Mkt_Mutex) );
      psMutex->poMutex    = objv[2]; _IRC( psMutex->poMutex );
      psMutex->poLockers  = _NOB   ; _IRC( psMutex->poLockers );

      _HASV( psHashE, psMutex );
      _SSO( _GOR, pcMutex );                           /* return mutex name */
      break;

    case _DELETE:                                         /* delete a mutex */
      _LOGL( psMutex->poLockers, &iIndex );   /* threads trying to lock it? */
      if( iIndex )                             /* if so, don't delete mutex */
        throw eStillUsed;

      _DRC( psMutex->poMutex );                       /* forget the objects */
      _DRC( psMutex->poLockers );
      _HADE( psHashE );

#ifdef MKT_NEWTHREADAPI
      Tcl_MutexFinalize( &psMutex->sMutex );                    /* clean-up */
      Tcl_ConditionFinalize( &psMutex->sCondition );
#else
      TclFinalizeMutex( &psMutex->sMutex );
      TclFinalizeCondition( &psMutex->sCondition );
#endif

      Tcl_Free( (char*)psMutex );                   /* release mutex struct */
      break;

    case _LOCK:                                             /* lock a mutex */
      if( psMutex->hOwner == psThread->hThread )
        throw eAlreadyLocked;        /* prevent crash from multiple locking */

      _LOAE( psMutex->poLockers, _NIO( (int)psThread->hThread ) );

      _MUNLOCK;
      Tcl_MutexLock( &psMutex->sMutex );
      _MLOCK;

      psMutex->hOwner = psThread->hThread;
      break;

    case _UNLOCK:                                         /* unlock a mutex */
      _MktRemoveFromList( psMutex->poLockers, psThread, &iIndex );
      if( iIndex == -1 )           /* prevent crash from multiple unlocking */
        throw eNotLocked;

      psMutex->hOwner = 0;
      Tcl_MutexUnlock( &psMutex->sMutex );
      break;

    case _WAIT:                     /* wait on a condition for notification */
      if( psMutex->hOwner != psThread->hThread )
        throw eNotLocked;             /* corresponding mutex must be locked */

      if( objc == 3 )                    /* no time specified: wait forever */
        psTime = NULL;
      else                              /* convert object to time structure */
        try( _MktGetTimeFromObj( pI, objv[3], psTime ), eError );

      psMutex->hOwner = 0;
      psMutex->bTimeout = TRUE;            /* notifier will set it to false */

      _MUNLOCK;
      Tcl_ConditionWait( &psMutex->sCondition, &psMutex->sMutex, psTime );
      _MLOCK;

      psMutex->hOwner = psThread->hThread;
      _SOR( _NIO( psMutex->bTimeout ) );
      break;

    case _NOTIFY:               /* notify a thread waiting on the condition */
      if( psMutex->sCondition )            /* if threads are waiting at all */
      {
        psMutex->hNotifier = psThread->hThread;
        psMutex->bTimeout = FALSE;
        Tcl_ConditionNotify( &psMutex->sCondition );          /* notify it! */
      }

      break;

    case _LOCKERS:       /* return list of threads trying to lock the mutex */
      _SOR( _DOB( psMutex->poLockers ) );
      break;

    case _NOTIFIER:                     /* return thread that last notified */
      _SOR( _NIO( (int)psMutex->hNotifier ) );
      break;

    case _OWNER:                                   /* return owner of mutex */
      _SOR( _NIO( (int)psMutex->hOwner ) );
      break;

    case _NAMES:                                  /* return all mutex names */
      for( _HASCAN( &tMutexes, psHashE, &sHashS ) )
        _LOAE( _GOR, _NSO( _HAGK( &tMutexes, psHashE ) ) );

      break;

    case _EXISTS:                         /* test if specified mutex exists */
      _SOR( _NIO( _HAFE( &tMutexes, _GSO( objv[2]) ) != NULL ) );
      break;
  }

  _MUNLOCK;

  return TCL_OK;

  catch eError:
    return TCL_ERROR;

  catch eBadMutex:
    _MUNLOCK;
    _APR( pI, "No such mutex '", _GSO( objv[2] ), "'", NULL );
    return TCL_ERROR;

  catch eMutexExists:
    _MUNLOCK;
    _APR( pI, "Mutex '", _GSO( objv[2] ), "' already exists", NULL );
    return TCL_ERROR;

  catch eAlreadyLocked:
    _MUNLOCK;
    _APR( pI, "Mutex '", _GSO( objv[2] ), "' already locked by this thread", NULL );
    return TCL_ERROR;

  catch eNotLocked:
    _MUNLOCK;
    _APR( pI, "Mutex '", _GSO( objv[2] ), "' not locked by this thread", NULL );
    return TCL_ERROR;

  catch eStillUsed:
    _MUNLOCK;
    _APR( pI, "Mutex '", _GSO( objv[2] ), "' still used", NULL );
    return TCL_ERROR;
}

/* Mkt_TvarCmd
implements the tvar command. since each thread must have a separate
interpreter, there is no standard way, e.g. by means of global variables,
to share values between threads. the tvar command, however, provides some
sort of variables (best compared with an array) which are common for all
threads. these thread variables can be set, retrieved, unset, and checked on
existance. a hash table is used, where the index is the tvar's name and the
value is a pointer to a Tcl_Obj containing the tvar's value. quite simple.
*/
int Mkt_TvarCmd( ClientData pC, Tcl_Interp *pI, int objc, Tcl_Obj *CONST objv[] )
{
  int             i, iOption, bNew;
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;

  static char *pArg[] = { "set", "get", "unset", "names", "exists", NULL };

  enum options { _SET, _GET, _UNSET, _NAMES, _EXISTS };

  if( objc < 2 )
    return _WNA( 2, "option ?args?" );

  try( _GIFO( objv[1], pArg, "option", &iOption ), eError );

  /* do some option checking up front */
  switch( iOption )
  {
    case _SET   : if( objc != 4 ) return _WNA( 2, "varName value" )     ; break;
    case _UNSET : if( objc < 3  ) return _WNA( 2, "varNname ?varName?" ); break;
    case _NAMES : if( objc != 2 ) return _WNA( 2, "" )                  ; break;
    default     : if( objc != 3 ) return _WNA( 2, "varName" )           ; break;
  }

  _MLOCK;

  switch( iOption )
  {
    case _SET:                                      /* create or set a tvar */
      psHashE = _HACE( &tTvars, _GSO( objv[2] ), &bNew );

      if( ! bNew )                       /* unreference any existing object */
        _DRC( (Tcl_Obj*)_HAGV( psHashE ) );

      _HASV( psHashE, (char*)objv[3] );                        /* set value */
      _IRC( objv[3] );                                 /* bump up ref count */
      break;

    case _GET:                                   /* retrieve a tvar's value */
      if( ( psHashE = _HAFE( &tTvars, _GSO( objv[2] ) ) ) == NULL )
        throw eNotFound;                  /* complain if tvar doesn't exist */

      _SOR( (Tcl_Obj*)_HAGV( psHashE ) );                      /* get value */
      break;

    case _UNSET:                                    /* forget about a tvars */
      for( i = 2; i < objc; i++ )        /* loop thru all given  tvar names */
      {
        if( ( psHashE = _HAFE( &tTvars, _GSO( objv[i] ) ) ) == NULL )
          throw eNotFound;                /* complain if tvar doesn't exist */

        _DRC( (Tcl_Obj*)_HAGV( psHashE ) );                /* forget object */
        _HADE( psHashE );                             /* delete table entry */
      }

      break;

    case _NAMES:                                      /* get all tvar names */
      for( _HASCAN( &tTvars, psHashE, &sHashS ) )
        _LOAE( _GOR, _NSO( _HAGK( &tTvars, psHashE ) ) );

      break;

    case _EXISTS:                           /* check on existance of a tvar */
      _SOR( _NIO( _HAFE( &tTvars, _GSO( objv[2] ) ) != NULL ) );
      break;
  }

  _MUNLOCK;

  return TCL_OK;

  catch eError:
    return TCL_ERROR;

  catch eNotFound:
    _MUNLOCK;
    _APR( pI, "Thread variable '", _GSO( objv[2] ), "' not defined.", NULL );
    return TCL_ERROR;
}

/* Mkt_SerializeCmd
implements the serialize command. a code segment is serialized by wrapping
a Tcl_Mutex around it. a serialized code segment is identifed by a name
(a key) which represents the used mutex. for each key a record in a hash
table is created, pointing to a Tcl_Mutex. this mutex is locked before
the given script is evaluated, and unlocked afterwards. if different code
segments use the same key, it is assured that no more than one thread is in
any of these segments at any time. if no key is provided, an empty string
is used as the key.
*/
int Mkt_SerializeCmd( ClientData pC, Tcl_Interp *pI, int objc, Tcl_Obj *CONST objv[] )
{
  int            iRes, bNew;
  Tcl_HashEntry *psHashE;
  Tcl_Mutex     *psMutex;

  if( objc < 2 )
    return _WNA( 1, "?key? script" );

  _MLOCK;

  psHashE = _HACE( &tSerials, (objc==2)? "":_GSO( objv[1] ), &bNew );
  if( bNew )                                          /* if this key is new */
  {             /* ...then allocate a Tcl_Mutex, initialize it and store it */
    psMutex = (Tcl_Mutex*)Tcl_Alloc( sizeof( Tcl_Mutex ) );
    *psMutex = 0;
    _HASV( psHashE, psMutex );
  }
  else
    psMutex = (Tcl_Mutex*)_HAGV( psHashE );      /* otherwise get its mutex */

  _MUNLOCK;

  Tcl_MutexLock( psMutex );                              /* lock this mutex */
  iRes = Tcl_EvalObj( pI, objv[objc-1] );             /* ...eval the script */
  Tcl_MutexUnlock( psMutex );                        /* and unlock it again */

  return iRes;
}

/* Mkthreads_Init, Mkthreads_SafeInit
package initialization. initializes the hash tables, creates a record
for the main thread and stores it in the threads table, then creates all
package commands. it also creates an init script for every new thread,
which is equivalent to a call to Tcl_Init(). this is necessary because
of a new bug in Tcl 8.3.1, where $tcl_library and $auto_path are not
set for new threads and therefore "package require" fails.
*/
int Mkthreads_Init( Tcl_Interp *pI )
{
  int bNew;

#ifdef USE_TCL_STUBS
  if( Tcl_InitStubs( pI, "8.3", 0) == NULL )
    throw eError;
#endif

  if( TCL_MAJOR_VERSION < 8 )
    throw eWrongVersion;

  /* check if tcl was compiled with TCL_THREADS defined */
  if( Tcl_GetVar2( pI, "tcl_platform", "threaded", TCL_GLOBAL_ONLY ) == NULL )
    throw eNoThreads;

  /* initialize hash tables */
  Tcl_InitHashTable( &tThreads , TCL_ONE_WORD_KEYS );
  Tcl_InitHashTable( &tPackages, TCL_STRING_KEYS );
  Tcl_InitHashTable( &tProcs   , TCL_STRING_KEYS );
  Tcl_InitHashTable( &tTvars   , TCL_STRING_KEYS );
  Tcl_InitHashTable( &tMutexes , TCL_STRING_KEYS );
  Tcl_InitHashTable( &tSerials , TCL_STRING_KEYS );

  /* initialize struct for main thread and put it into threads table */
  sMainThread.hThread  = Tcl_GetCurrentThread();
  sMainThread.psInterp = pI;
  sMainThread.psEntry  = _HACE( &tThreads, (char*)sMainThread.hThread, &bNew );
  _HASV( sMainThread.psEntry, &sMainThread );

  /* create init script to emulate Tcl_Init() */
  poInitScript = _NOB;
  _ASO( poInitScript, "if { ! [info exists tcl_library] } { set tcl_library {", Tcl_GetVar( pI, "tcl_library", TCL_GLOBAL_ONLY ), 
        "}; source $tcl_library/init.tcl; set auto_path {", 
        Tcl_GetVar( pI, "auto_path", TCL_GLOBAL_ONLY ), "} }", NULL );   
  _IRC( poInitScript );

  /* setup unload handler and create new commands */
  Tcl_CallWhenDeleted( pI, Mkthreads_Exit, NULL );
  _MktCreateCommands( &sMainThread );

  try( Tcl_PkgProvide( pI, "mkThreads", _VERSION ), eError );

  return TCL_OK;

  catch eError:
    return TCL_ERROR;

  catch eWrongVersion:
    _SSO( _GOR, "Package mkThreads requires Tcl version 8" );
    return TCL_ERROR;

  catch eNoThreads:
    _SSO( _GOR, "No thread support with this Tcl version." );
    return TCL_ERROR;
}

int Mkthreads_SafeInit( Tcl_Interp *pI )
{
  return Mkthreads_Init( pI );
}

/* Mkthreads_Exit
termination procedure. called when the package is unloaded. frees up any
memory previously allocated. warning: should not be called as long as
threads exist.
*/
void Mkthreads_Exit( ClientData pC, Tcl_Interp *pI )
{
  Tcl_HashEntry  *psHashE;
  Tcl_HashSearch  sHashS;
  Mkt_Thread *psThread;
  Mkt_Mutex  *psMutex;

  _MLOCK;

  for( _HASCAN( &tThreads, psHashE, &sHashS ) )
  {
    psThread = (Mkt_Thread*)_HAGV( psHashE );
    if( psThread == &sMainThread )
      continue;

    Tcl_DeleteInterp( psThread->psInterp );
    Tcl_Free( (char*)psThread );
  }

  for( _HASCAN( &tMutexes, psHashE, &sHashS ) )
  {
    psMutex = (Mkt_Mutex*)_HAGV( psHashE );
    _DRC( psMutex->poMutex );
    _DRC( psMutex->poLockers );
    Tcl_Free( (char*)psMutex );
  }

  for( _HASCAN( &tPackages, psHashE, &sHashS ) )
    _DRC( (Tcl_Obj*)_HAGV( psHashE ) );

  for( _HASCAN( &tProcs, psHashE, &sHashS ) )
    _DRC( (Tcl_Obj*)_HAGV( psHashE ) );

  for( _HASCAN( &tTvars, psHashE, &sHashS ) )
    _DRC( (Tcl_Obj*)_HAGV( psHashE ) );

  for( _HASCAN( &tSerials, psHashE, &sHashS ) )
    Tcl_Free( (char*)_HAGV( psHashE ) );

  if( sOnexit.poScript )
    _DRC( sOnexit.poScript );

  Tcl_DeleteHashTable( &tThreads );
  Tcl_DeleteHashTable( &tMutexes );
  Tcl_DeleteHashTable( &tPackages );
  Tcl_DeleteHashTable( &tProcs );
  Tcl_DeleteHashTable( &tTvars );
  Tcl_DeleteHashTable( &tSerials );

  if( poInitScript )
    _DRC( poInitScript );

  _MUNLOCK;
}

/* static linking. uncomment the following two functions if you want
to create a stand-alone shell instead of a dynamic library. */

#ifndef USE_TCL_STUBS
int main( int argc, char *argv[] )
{
  Tcl_Main( argc, argv, Tcl_AppInit );
  return 0;
}

int Tcl_AppInit( Tcl_Interp *pI )
{
  try( Tcl_Init( pI ), eError );
  try( Mkthreads_Init( pI ), eError );

  return TCL_OK;

  catch eError:
    return TCL_ERROR;
}
#endif

/*
 * mkThreads 1.2
 * -------------
 */

